//--------------------------------------------------------------------------------------
//	$Id: u32p32exp2m1.cc,v 1.1 2024/03/13 20:18:38 wolf Exp $
//--------------------------------------------------------------------------------------

#include <stdint.h>

#include "u32p32exp2m1.h"

//--------------------------------------------------------------------------------------
//	aux functions
//--------------------------------------------------------------------------------------

inline constexpr uint32_t
u32mulhi( const uint32_t a, const uint32_t b ){

	return (uint32_t) (((uint64_t) a * b) >> 32);
}

//--------------------------------------------------------------------------------------
//	Integerberechung von u32bexpm1(x) := 2**32  *  ( 2 ** (x/2**32) - 1 )
//--------------------------------------------------------------------------------------

// pow(2, 32) / 3
// pow(2, 32) / 5
// pow(2, 32) / 6
// pow(2, 32) / 7

static constexpr uint32_t ONE_BY_3 = (uint32_t)( ( 1ul << 32 ) / 3 );
static constexpr uint32_t ONE_BY_5 = (uint32_t)( ( 1ul << 32 ) / 5 );
static constexpr uint32_t ONE_BY_6 = (uint32_t)( ( 1ul << 32 ) / 6 );
static constexpr uint32_t ONE_BY_7 = (uint32_t)( ( 1ul << 32 ) / 7 );

// pow(2, 32) * log(2) - unfortunately log(const) is not a constexpr

static constexpr uint32_t LOG_2 = 0xb17217f7u;

// two top bits are handled by lookup table,
// resulting in input of series expansion to be less then 0.25

static constexpr unsigned WORD_LEN   = 32;
static constexpr unsigned INT_BITS   =  2;
static constexpr unsigned SHIFT_BITS = WORD_LEN - INT_BITS;

// unfortunately exp() is not a constexpr

// exp( 1/4 ) * 2**32 = 0x1.48b5.e3c3

static constexpr uint32_t EXP_1_BY_4_INT  =  1u;
static constexpr uint32_t EXP_1_BY_4_FRAC = 0x48b5e3c3u;

// exp( 2/4 ) * 2**32 = 0x1.a612.98e1

static constexpr uint32_t EXP_2_BY_4_INT  =  1u;
static constexpr uint32_t EXP_2_BY_4_FRAC = 0xa61298e1u;

// exp( 3/4 ) * 2**32 = 0x2.1df3.b68c [not needed]

static constexpr uint32_t EXP_3_BY_4_INT  =  2u;
static constexpr uint32_t EXP_3_BY_4_FRAC = 0x1df3b68cu;

//--------------------------------------------------------------------------------------
//	definition
//--------------------------------------------------------------------------------------

#include "u32p32exp2m1.h"

//--------------------------------------------------------------------------------------
//	implementation
//--------------------------------------------------------------------------------------

uint32_t
u32p32exp2m1( uint32_t arg ){

	//------------------------------------------------------------------------------
	//	exp2( arg ) = exp( arg * log(2) )
	//------------------------------------------------------------------------------
	//	exp2( arg ) = exp(     x        )		x := arg * log(2)
	//------------------------------------------------------------------------------

	const uint32_t x = u32mulhi( arg, LOG_2 );

	//------------------------------------------------------------------------------
	//	separate leading two bits (handled by table lookup)
	//------------------------------------------------------------------------------
	//	exp( x ) = exp( x_int << 30 ) * exp( x_fract )
	//------------------------------------------------------------------------------

	const uint32_t x_int  = x          >> SHIFT_BITS;
	const uint32_t x_frac = x - (x_int << SHIFT_BITS);

	//------------------------------------------------------------------------------
	//	compute series expansion with modified horner scheme
	//------------------------------------------------------------------------------

	//------------------------------------------------------------------------------
	//	temp := x
	//------------------------------------------------------------------------------

	uint32_t temp = x_frac;

	//------------------------------------------------------------------------------
	//	temp := x + x/7 * x
	//------------------------------------------------------------------------------

	temp = u32mulhi( u32mulhi( temp, x_frac ), ONE_BY_7 ) + x_frac;

	//------------------------------------------------------------------------------
	//	temp := x + x/6 * (x + x/7 * x)
	//------------------------------------------------------------------------------

	temp = u32mulhi( u32mulhi( temp, x_frac ), ONE_BY_6 ) + x_frac;

	//------------------------------------------------------------------------------
	//	temp := x + x/5 * (x + x/6 * (x + x/7 * x))
	//------------------------------------------------------------------------------

	temp = u32mulhi( u32mulhi( temp, x_frac ), ONE_BY_5 ) + x_frac;

	//------------------------------------------------------------------------------
	//	temp := x + x/4 * (x + x/5 * (x + x/6 * (x + x/7 * x)))
	//------------------------------------------------------------------------------

	temp = ( u32mulhi( temp, x_frac ) >> 2) + x_frac;

	//------------------------------------------------------------------------------
	//	temp := x + x/3 * (x + x/4 * (x + x/5 * (x + x/6 * (x + x/7 * x))))
	//------------------------------------------------------------------------------

	temp = u32mulhi( u32mulhi( temp, x_frac ), ONE_BY_3 ) + x_frac;

	//------------------------------------------------------------------------------
	// temp := x + x/2 * (x + x/3 * (x + x/4 * (x + x/5 * (x + x/6 * (x + x/7 * x)))))
	//------------------------------------------------------------------------------

	temp = ( u32mulhi( temp, x_frac ) >> 1) + x_frac;

	//------------------------------------------------------------------------------
	// NOW: temp = exp(x_fract)-1 = x + x²/2! + x³/3! + x⁴/4! + x⁵/5! + x⁶/6! + x⁷/7!
	//------------------------------------------------------------------------------
	//	note the "-1" resulting from the missing "1" in the series expansion
	//------------------------------------------------------------------------------

	//------------------------------------------------------------------------------
	//	multiply exp(x_int) by exp( int_frac)
	//------------------------------------------------------------------------------
	//	exp( x_int  ) = exp( n/4 ) = 2^32 * EXP_n_BY_4_INT + EXP_n_BY_4_FRAC
	//------------------------------------------------------------------------------
	//	exp( x_frac ) = 2^32 + temp			| 2^32 is missing 1
	//------------------------------------------------------------------------------
	//	exp( x ) = exp( x_int ) * exp( x_fract )	| distribute
	//------------------------------------------------------------------------------
	//	exp(x)	=        EXP_n_BY_4_FRAC * temp	| u64mul mit shift
	//		+ 2^32 * EXP_n_BY_4_INT  * temp	| u32mul ohne Shift (wg. 2^32)
	//		+        EXP_n_BY_4_FRAC * 2^32	| copy              (wg. 2^32)
	//		+ 2^32 * EXP_n_BY_4_INT  * 2^32	| delete wg. 2^64
	//------------------------------------------------------------------------------

	switch( x_int ){

		//----------------------------------------------------------------------
		//	x_int = 0: multiply by exp( 0 ) == 1 --> do nothing
		//----------------------------------------------------------------------
	default:
		break;

		//----------------------------------------------------------------------
		//	x_int = 1: multiply by exp( 1/4 ) 
		//----------------------------------------------------------------------
		//	exp( 1/4 ) = 1 + EXP_1_BY_4_INT, where EXP_1_BY_4_INT < 1
		//----------------------------------------------------------------------

	case 1: temp =	u32mulhi( EXP_1_BY_4_FRAC, temp ) +
			EXP_1_BY_4_INT * temp + EXP_1_BY_4_FRAC;
		break;

		//----------------------------------------------------------------------
		//	x_int = 2: multiply by exp( 2/4 ) 
		//----------------------------------------------------------------------
		//	exp( 2/4 ) = 1 + EXP_2_BY_4_INT, where EXP_2_BY_4_INT < 1
		//----------------------------------------------------------------------

	case 2: temp =	u32mulhi( EXP_2_BY_4_FRAC, temp ) +
			EXP_2_BY_4_INT * temp + EXP_2_BY_4_FRAC;
		break;

		//----------------------------------------------------------------------
		//	x_int = 3: multiply by exp( 3/4 ) 		[unneeded]
		//----------------------------------------------------------------------
		//	exp( 3/4 ) = 1 + EXP_3_BY_4_INT, where EXP_3_BY_4_INT < 1
		//----------------------------------------------------------------------

	case 3: temp =	u32mulhi( EXP_3_BY_4_FRAC, temp ) +
			EXP_3_BY_4_INT * temp + EXP_3_BY_4_FRAC;
	}

	// heuristic value correcting truncation and series remainder

	return temp + ( u32mulhi( arg, 12 ) >> 1 );
}

//--------------------------------------------------------------------------------------
//	$Id: u32p32exp2m1.cc,v 1.1 2024/03/13 20:18:38 wolf Exp $
//--------------------------------------------------------------------------------------
